@@ -25,6 +25,7 @@
 
 #include <comphelper/processfactory.hxx>
 #include <com/sun/star/lang/WrappedTargetRuntimeException.hpp>
+#include <com/sun/star/lang/IllegalArgumentException.hpp>
 #include <com/sun/star/ucb/CommandAbortedException.hpp>
 #include <com/sun/star/ucb/CommandFailedException.hpp>
 #include <com/sun/star/uno/Sequence.hxx>
@@ -39,10 +40,8 @@
 #include <com/sun/star/xml/sax/XSAXSerializable.hpp>
 #include <com/sun/star/xml/sax/Writer.hpp>
 #include <com/sun/star/xml/sax/XWriter.hpp>
-#include <com/sun/star/io/XStream.hpp>
 #include <com/sun/star/io/TempFile.hpp>
 #include <com/sun/star/io/XOutputStream.hpp>
-#include <com/sun/star/beans/XPropertySet.hpp>
 #include <cppuhelper/exc_hlp.hxx>
 
 using namespace comphelper;
@@ -707,8 +706,7 @@ namespace
                 return false;
             }
 
-            // coverity#1373663 Untrusted loop bound, check file size
-            // isn't utterly broken
+            // Check that file size isn't utterly broken
             sal_uInt64 nFileSize(0);
             rFile->getSize(nFileSize);
             if (nFileSize < nExtEntries)
@@ -1185,7 +1183,7 @@ namespace
                                         if (aEntry.read_header(aSourceFile))
                                         {
                                             // add to local data
-                                            maPackedFileEntryVector.push_back(aEntry);
+                                            maPackedFileEntryVector.push_back(std::move(aEntry));
                                         }
                                         else
                                         {
@@ -1449,7 +1447,6 @@ namespace
 
 namespace comphelper
 {
-    sal_uInt16 BackupFileHelper::mnMaxAllowedBackups = 10;
     bool BackupFileHelper::mbExitWasCalled = false;
     bool BackupFileHelper::mbSafeModeDirExists = false;
     OUString BackupFileHelper::maInitialBaseURL;
